#include <ruby.h>
#include <rubyio.h>
#include <attr/attributes.h>
#include <attr/xattr.h>

/*
 *  call-seq:
 *      File.list_attrs(filename) -> array
 *
 *  Returns an array with all attributes
 *
 *      File.list_attrs(".") => ("type", "rating", "artist")
 *
 */

static VALUE rb_file_list_attrs(obj, fname)
	VALUE obj, fname;
{
	int size;
	char *lval;
	int lret;
	int beg, end;
	VALUE retval;

	Check_Type(fname, T_STRING);

	size  = 16000;
	lval   = malloc(size);

	lret = llistxattr(StringValueCStr(fname), lval, size);
	if(lret >= 0) {
		retval = rb_ary_new();
		beg = 0;
		for(end=0; end < lret; end++) {
			if (beg == 0 && lval[end] == '.') {
				beg = end + 1;
			}
			if (lval[end] == 0) {
				rb_ary_push(retval,
					    rb_str_new(lval+beg, end-beg));
				beg = 0;
			}
		}
		free(lval);
		return retval;
	} else {
		free(lval);
		rb_sys_fail(StringValueCStr(fname));
	}
}

/*
 *  call-seq:
 *      file.list_attrs() -> array
 *
 *  Returns an array with all attributes
 *
 *      File.new(__FILE__).list_attrs() => ("type", "rating", "artist")
 *
 */

VALUE rb_file_list_attrsf(fobj)
	VALUE fobj;
{
	int size;
	char *lval;
	int lret;
	int beg, end;
	VALUE retval;
	OpenFile *fptr;
	FILE *f;

	GetOpenFile(fobj, fptr);
	f = GetReadFile(fptr);

	Check_Type(fobj, T_FILE);

	size   = 16000;
	lval   = malloc(size);

	lret = flistxattr(fileno(f), lval, size);
	if(lret >= 0) {
		retval = rb_ary_new();
		beg = 0;
		for(end=0; end < lret; end++) {
			if (beg == 0 && lval[end] == '.') {
				beg = end + 1;
			}
			if (lval[end] == 0) {
				rb_ary_push(retval,
					    rb_str_new(lval+beg, end-beg));
				beg = 0;
			}
		}
		free(lval);
		return retval;
	} else {
		free(lval);
		rb_sys_fail("");
	}
}

/*
 *  call-seq:
 *      File.get_all_attrs(filename) -> hash
 *
 *  Returns a hash with all attributes and their values
 *
 *      File.get_all_attrs(".") => ("mime" => "text/plain", "rating" => "4")
 *
 */

static VALUE rb_file_get_all_attrs(obj, fname)
	VALUE obj, fname;
{
	int size;
	char *lval;
	char *gval;
	int *gsize;
	int lret, gret;
	int beg, end;
	VALUE retval;

	Check_Type(fname, T_STRING);

	size  = 16000;
	lval   = malloc(size);
	gsize  = malloc(sizeof(int));
	gval   = malloc(size);

	lret = llistxattr(StringValueCStr(fname), lval, size);
	if(lret >= 0) {
		retval = rb_hash_new();
		beg = 0;
		for(end=0; end < lret; end++) {
			if (beg == 0 && lval[end] == '.') {
				beg = end + 1;
			}
			if (lval[end] == 0) {
				*gsize = size;
				gret = attr_get(StringValueCStr(fname),
						lval+beg, gval, gsize, 1);
				if(gret == 0) {
					rb_hash_aset(retval,
						     rb_str_new(lval+beg,
								end-beg),
						     rb_str_new(gval, *gsize));
				} else {
					if(errno != ENODATA && errno != ENOATTR) {
						free(lval);
						free(gval);
						free(gsize);
						rb_sys_fail(StringValueCStr(fname));
					}
				}
				beg = 0;
			}
		}
		
		free(lval);
		free(gval);
		free(gsize);
		return retval;
	} else {
		free(lval);
		free(gval);
		free(gsize);
		rb_sys_fail(StringValueCStr(fname));
	}
}

/*
 *  call-seq:
 *  	file.get_all_attrs() -> hash
 *
 *  Returns a hash with all attributes and their values
 *
 *  	File.new(__FILE__).get_all_attrs() => ("type" => "jpeg", "rat" => "4")
 *
 */

VALUE rb_file_get_all_attrsf(fobj)
	VALUE fobj;
{
	int size;
	char *lval;
	char *gval;
	int *gsize;
	int lret, gret;
	int beg, end;
	VALUE retval;
	OpenFile *fptr;
	FILE *f;

	GetOpenFile(fobj, fptr);
	f = GetReadFile(fptr);

	Check_Type(fobj, T_FILE);

	size   = 16000;
	lval   = malloc(size);
	gsize  = malloc(sizeof(int));
	gval   = malloc(size);

	lret = flistxattr(fileno(f), lval, size);
	if(lret >= 0) {
		retval = rb_hash_new();
		beg = 0;
		for(end=0; end < lret; end++) {
			if (beg == 0 && lval[end] == '.') {
				beg = end+1;
			}
			if (lval[end] == 0) {
				*gsize = size;
				gret = attr_getf(fileno(f),
						 lval+beg, gval, gsize, 1);
				if(gret == 0) {
					rb_hash_aset(retval,
						     rb_str_new(lval+beg,
								end-beg),
						     rb_str_new(gval, *gsize));
				} else {
					if(errno != ENODATA && errno != ENOATTR) {
						free(lval);
						free(gval);
						free(gsize);
						rb_sys_fail("");
					}
				}
				beg = 0;
			}
		}
		
		free(lval);
		free(gval);
		free(gsize);
		return retval;
	} else {
		free(lval);
		free(gval);
		free(gsize);
		rb_sys_fail("");
	}
}
  

/*
 *  call-seq:
 *      File.get_attr(filename, attribute) -> attribute_value
 *
 *  Returns the value of the attribute <i>attribute</i>
 *
 *      File.get_attr(".", "description") => "ruby-xattr directory and files"
 *
 */

static VALUE rb_file_get_attr(obj, fname, aname)
	VALUE obj, fname, aname;
{
	char *aval;
	int *size;
	int ret;
	VALUE retval;

	Check_Type(fname, T_STRING);
	Check_Type(aname, T_STRING);

	size = malloc(sizeof(int));
	aval = malloc(16000);
	*size = 16000;

	ret = attr_get(StringValueCStr(fname), StringValueCStr(aname), aval, size, 1);
	if(ret == 0) {
		retval = rb_str_new(aval, *size);
		free(aval);
		free(size);
		return retval;
	} else {
		free(aval);
		free(size);
		if(errno != ENOATTR && errno != ENODATA) {
			rb_sys_fail(StringValueCStr(fname));
		}
		return Qnil;
	}
}

/*
 *  call-seq:
 *  	file.get_attr(attribute) -> attribute_value
 *
 *  Returns the value of <i>attribute</i> for the file
 *
 *  	File.new(__FILE__).get_attr("description") # => "test script"
 *
 */

VALUE rb_file_get_attrf(fobj, aname)
	VALUE fobj, aname;
{
	char *aval;
	int *size;
	int ret;
	VALUE retval;
	OpenFile *fptr;
	FILE *f;

	GetOpenFile(fobj, fptr);
	f = GetReadFile(fptr);

	Check_Type(fobj, T_FILE);
	Check_Type(aname, T_STRING);

	size = malloc(sizeof(int));
	aval = malloc(16000);
	*size = 16000;

	ret = attr_getf(fileno(f), StringValueCStr(aname), aval, size, 1);
	if(ret == 0) {
		retval = rb_str_new(aval, *size);
		free(aval);
		free(size);
		return retval;
	} else {
		free(aval);
		free(size);
		if(ret != ENODATA && ret != ENOATTR) {
			rb_sys_fail("");
		}
		return Qnil;
	}
}

/*
 *  call-seq:
 *  	File.set_attr(filename, attribute, attribute_value) -> attribute_value
 *
 *  Sets the attribute <i>attribute</i> for the file <i>filename</i>
 *
 *  	File.set_attr("test.rb", "description", "test script") # => "test script"
 *
 */


static VALUE rb_file_set_attr(obj, fname, aname, aval)
	VALUE obj, fname, aname, aval;
{
	int ret;

	Check_Type(fname, T_STRING);
	Check_Type(aname, T_STRING);
	Check_Type(aval, T_STRING);

	ret = attr_set(StringValueCStr(fname), StringValueCStr(aname), 
			StringValueCStr(aval), RSTRING(aval)->len, 1);
	if(ret == 0) {
		return aval;
	} else {
		rb_sys_fail(StringValueCStr(fname));
	}
}

/*
 *  call-seq:
 *  	file.set_attr(attribute, attribute_value) -> attribute_value
 *
 *  Sets the attribute <i>attribute</i> for the file.
 *
 *  	File.new("test.rb").set_attr("description", "test script") # => "test script"
 *
 */

VALUE rb_file_set_attrf(fobj, aname, aval)
	  VALUE fobj, aname, aval;
{
	int ret;
	OpenFile *fptr;
	FILE *f;

	GetOpenFile(fobj, fptr);
	f = GetReadFile(fptr);

	Check_Type(fobj, T_FILE);
	Check_Type(aname, T_STRING);
	Check_Type(aval, T_STRING);
	ret = attr_setf(fileno(f), StringValueCStr(aname), 
			StringValueCStr(aval), RSTRING(aval)->len, 1);
	if(ret == 0) {
		return aval;
	} else {
		rb_sys_fail("");
	}
}

/*
 *  call-seq:
 *      File.remove_attr(filename, attribute)
 *
 *  Removes <i>attribute</i> from the given file
 *
 *      File.remove_attr(".", "description") 
 *
 */

static VALUE rb_file_remove_attr(obj, fname, aname)
        VALUE obj, fname, aname;
{
	int ret;

	Check_Type(fname, T_STRING);
	Check_Type(aname, T_STRING);

	ret = attr_remove(StringValueCStr(fname), StringValueCStr(aname), 1);
	if (ret != 0)
		rb_sys_fail(StringValueCStr(fname));
	return 1;
}

/*
 *  call-seq:
 *  	file.remove_attr(attribute)
 *
 *  Removes <i>attribute</i> from the file
 *
 *  	File.new(__FILE__).remove_attr("description")
 *
 */

VALUE rb_file_remove_attrf(fobj, aname)
	VALUE fobj, aname;
{
	int ret;
	OpenFile *fptr;
	FILE *f;

	GetOpenFile(fobj, fptr);
	f = GetReadFile(fptr);

	Check_Type(fobj, T_FILE);
	Check_Type(aname, T_STRING);

	ret = attr_removef(fileno(f), StringValueCStr(aname), 1);
	if (ret != 0)
		rb_sys_fail("");
	return 1;
}

void Init_xattr () {
	rb_define_singleton_method(rb_cFile, "list_attrs",
				   rb_file_list_attrs, 1);
	rb_define_singleton_method(rb_cFile, "get_all_attrs",
				   rb_file_get_all_attrs, 1);
	rb_define_singleton_method(rb_cFile, "get_attr", rb_file_get_attr, 2);
	rb_define_singleton_method(rb_cFile, "set_attr", rb_file_set_attr, 3);
	rb_define_singleton_method(rb_cFile, "remove_attr",
				   rb_file_remove_attr, 2);
	rb_define_method(rb_cFile, "list_attrs", rb_file_list_attrsf, 0);
	rb_define_method(rb_cFile, "get_all_attrs", rb_file_get_all_attrsf, 0);
	rb_define_method(rb_cFile, "get_attr", rb_file_get_attrf, 1);
	rb_define_method(rb_cFile, "set_attr", rb_file_set_attrf, 2);
	rb_define_method(rb_cFile, "remove_attr", rb_file_remove_attrf, 1);
}
